from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, Literal, TypedDict

from mcp.server.fastmcp.server import Context


class BaseExecutor(ABC):
    @abstractmethod
    def submit(self, fn: Callable, kwargs: dict) -> TypedDict(
            'results', {'job_id': str, 'extra_info': dict}):
        pass

    @abstractmethod
    def query_status(self, job_id: str) -> Literal[
            "Running", "Succeeded", "Failed"]:
        pass

    @abstractmethod
    def terminate(self, job_id: str) -> None:
        pass

    @abstractmethod
    def get_results(self, job_id: str) -> dict:
        pass

    @abstractmethod
    async def async_run(
        self, fn: Callable, kwargs: dict, context: Context) -> TypedDict(
            'results', {'job_id': str, 'extra_info': dict, 'result': Any}):
        pass



def merge(d1, d2):
    for k in d2.keys():
        if k in d1 and isinstance(d1[k], dict) and isinstance(d2[k], dict):
            merge(d1[k], d2[k])
        else:
            d1[k] = d2[k]


def bohrium_preprocessor(default_image, default_machine_type):
    def preprocess(executor, storage, kwargs):
        modified_executor = {
            "type": "dispatcher",
            "machine": {
                "batch_type": "Bohrium",
                "context_type": "Bohrium",
                "remote_profile": {
                    "input_data": {
                        "image_name": default_image,
                        "job_type": "container",
                        "platform": "ali",
                        "scass_type": default_machine_type,
                    },
                },
            },
        }
        if executor and executor.get("type") == "dispatcher" and executor.get(
                "machine", {}).get("context_type") == "Bohrium":
            merge(modified_executor, executor)
        modified_storage = {
            "type": "bohrium",
        }
        if storage and storage.get("type") == "bohrium":
            merge(modified_storage, storage)
        return modified_executor, modified_storage, kwargs
    return preprocess



from .client import CalculationMCPTool, CalculationMCPToolset
from .utils import update_session_handler

__all__ = ["CalculationMCPTool", "CalculationMCPToolset",
           "update_session_handler"]



import jsonpickle
import time
from copy import deepcopy
from typing import Any, Dict, Optional

from google.adk.events import Event, EventActions
from google.adk.tools.base_tool import BaseTool
from google.adk.tools.tool_context import ToolContext


def update_session_handler(session_service, session):
    """Update session with job and artifact information after tool calling."""
    async def after_tool_handler(
        tool: BaseTool, args: Dict[str, Any], tool_context: ToolContext,
        tool_response: dict,
    ) -> Optional[Dict]:
        results = jsonpickle.loads(tool_response.content[0].text)
        jobs = session.state.get("jobs", [])
        results["tool_name"] = tool.name
        user_args = deepcopy(args)
        user_args.pop("executor", {})
        user_args.pop("storage", {})
        results["args"] = user_args
        results["agent_name"] = tool_context.agent_name
        jobs.append(results)
        artifacts = session.state.get("artifacts", [])
        artifacts = {art["uri"]: art for art in artifacts}
        for name, art in results["input_artifacts"].items():
            if art["uri"] not in artifacts:
                artifacts[art["uri"]] = {
                    "type": "input",
                    "name": name,
                    "job_id": results["job_id"],
                    **art,
                }
        for name, art in results["output_artifacts"].items():
            if art["uri"] not in artifacts:
                artifacts[art["uri"]] = {
                    "type": "output",
                    "name": name,
                    "job_id": results["job_id"],
                    **art,
                }
        artifacts = list(artifacts.values())
        state_changes = {
            "jobs": jobs,
            "artifacts": artifacts,
        }
        actions_with_update = EventActions(state_delta=state_changes)
        system_event = Event(
            invocation_id="inv_login_update",
            author="system",
            actions=actions_with_update,
            timestamp=time.time()
        )
        await session_service.append_event(session, system_event)
        return None
    return after_tool_handler



import os
import tarfile
from abc import ABC, abstractmethod
from typing import List


class BaseStorage(ABC):
    @abstractmethod
    def _upload(self, key: str, path: str) -> str:
        """
        Upload a file from path to key
        """
        pass

    @abstractmethod
    def _download(self, key: str, path: str) -> str:
        """
        Download a file from key to path
        """
        pass

    @abstractmethod
    def list(self, prefix: str, recursive: bool = False) -> List[str]:
        pass

    @abstractmethod
    def copy(self, src: str, dst: str) -> None:
        pass

    @abstractmethod
    def get_md5(self, key: str) -> str:
        pass

    def download(self, key: str, path: str) -> str:
        objs = self.list(prefix=key, recursive=True)
        if objs == [key]:
            path = os.path.join(path, os.path.basename(key))
            self._download(key=key, path=path)
            if path[-4:] == ".tgz":
                path = extract(path)
        else:
            for obj in objs:
                rel_path = obj[len(key):]
                if rel_path[:1] == "/":
                    rel_path = rel_path[1:]
                file_path = os.path.join(path, rel_path)
                self._download(key=obj, path=file_path)
        return path

    def upload(self, key: str, path: str) -> str:
        if os.path.isfile(path):
            key = os.path.join(key, os.path.basename(path))
            key = self._upload(key, path)
        elif os.path.isdir(path):
            cwd = os.getcwd()
            if os.path.dirname(path):
                os.chdir(os.path.dirname(path))
            fname = os.path.basename(path)
            with tarfile.open(fname + ".tgz", "w:gz", dereference=True) as tf:
                tf.add(fname)
            os.chdir(cwd)
            key = os.path.join(key, fname + ".tgz")
            key = self._upload(key, "%s.tgz" % path)
            os.remove("%s.tgz" % path)
        return key


def extract(path):
    with tarfile.open(path, "r:gz") as tf:
        common = os.path.commonpath(tf.getnames())
        tf.extractall(os.path.dirname(path))

    os.remove(path)
    path = os.path.dirname(path)
    # if the tarfile contains only one directory,
    # return its path
    if common != "":
        return os.path.join(path, common)
    else:
        return path



import logging
from typing import List, Optional

from mcp import ClientSession, types
from google.adk.tools.mcp_tool import MCPTool, MCPToolset
from google.adk.tools.mcp_tool.mcp_session_manager import MCPSessionManager
logger = logging.getLogger(__name__)


async def logging_handler(
    params: types.LoggingMessageNotificationParams,
) -> None:
    logger.log(getattr(logging, params.level.upper()), params.data)


class MCPSessionManagerWithLoggingCallback(MCPSessionManager):
    def __init__(
      self,
      logging_callback=None,
      **kwargs,
    ):
        super().__init__(**kwargs)
        self.logging_callback = logging_callback

    async def create_session(self) -> ClientSession:
        session = await super().create_session()
        session._logging_callback = self.logging_callback
        return session


class CalculationMCPTool(MCPTool):
    def __init__(
        self,
        executor: Optional[dict] = None,
        storage: Optional[dict] = None,
    ):
        """Calculation MCP tool
        extended from google.adk.tools.mcp_tool.MCPTool

        Args:
            executor: The executor configuration of the calculation tool.
                It is a dict where the "type" field specifies the executor
                type, and other fields are the keyword arguments of the
                corresponding executor type.
            storage: The storage configuration for storing artifacts. It is
                a dict where the "type" field specifies the storage type,
                and other fields are the keyword arguments of the
                corresponding storage type.
        """
        self.executor = executor
        self.storage = storage

    async def run_async(self, args, **kwargs):
        if "executor" not in args:
            args["executor"] = self.executor
        if "storage" not in args:
            args["storage"] = self.storage
        return await super().run_async(args=args, **kwargs)


class CalculationMCPToolset(MCPToolset):
    def __init__(
        self,
        executor: Optional[dict] = None,
        storage: Optional[dict] = None,
        executor_map: Optional[dict] = None,
        **kwargs,
    ):
        """
        Calculation MCP toolset

        Args:
            executor: The default executor configuration of the calculation
                tools. It is a dict where the "type" field specifies the
                executor type, and other fields are the keyword arguments of
                the corresponding executor type.
            storage: The storage configuration for storing artifacts. It is
                a dict where the "type" field specifies the storage type,
                and other fields are the keyword arguments of the
                corresponding storage type.
            executor_map: A dict mapping from tool name to executor
                configuration for specifying particular executor for certain
                tools
        """
        super().__init__(**kwargs)
        self._mcp_session_manager = MCPSessionManagerWithLoggingCallback(
            connection_params=self._connection_params,
            errlog=self._errlog,
            logging_callback=logging_handler,
        )
        self.executor = executor
        self.storage = storage
        self.executor_map = executor_map or {}

    async def get_tools(self, *args, **kwargs) -> List[CalculationMCPTool]:
        tools = await super().get_tools(*args, **kwargs)
        calc_tools = []
        for tool in tools:
            calc_tool = CalculationMCPTool(
                executor=self.executor_map.get(tool.name, self.executor),
                storage=self.storage)
            calc_tool.__dict__.update(tool.__dict__)
            calc_tools.append(calc_tool)
        return calc_tools
